/*########################################################################################################################################################
 *
 *  get entries of table linecat
 *  Copyright (C) 2012 - 2024  Thomas Moeller
 *
 *  I. Physikalisches Institut, University of Cologne
 *
 *
 *
 *
 *
 *  input variables:     NumberMoleculeRadTrans:    number of molecules
 *                       querylength:               length of query
 *                       queryString1:              first part of query string
 *                       queryString2:              second part of query string
 *                       NumberOfMolecules:         number of molecules which should be considered (used for query string)
 *                       MoleculeNames:             names of molecules which should be considered (used for query string)
 *                       dbName:                    path and name database file
 *                       NameOfColumnName:          name of column containing the name
 *
 *  output variables:    ok:                        status variable
 *                       MolNameRadTrans:           names of the molecules in database
 *                       lFreq:                     frequency of the transition in MHz
 *                       lFreqErr:                  corresponding uncertainty in MHz
 *                       icatMol:                   base 10 logarithm of the integrated intensity
 *                       EinsteinA:                 Einstein A coefficients
 *                       lElt:                      energy of lower level
 *                       UpperStateDegeneracy:      upper state degeneracy
 *                       LowerQNString:             quantum number string of lower state
 *                       UpperQNString:             quantum number string of upper state
 *
 *
 *
 *  Versions of the program:
 *
 *  Who           When        What
 *  T. Moeller    2012-05-31  Initial version
 *  T. Moeller    2013-01-22  change from MySQL to sqlite3
 *
 *
 *
 *  License:
 *
 *    GNU GENERAL PUBLIC LICENSE
 *    Version 3, 29 June 2007
 *    (Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>)
 *
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
########################################################################################################################################################*/


/* make a connection to the database */
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sqlite3.h>
#include <time.h>
#include <math.h>
#include <sys/times.h>

#define nameLength 60
#define QNStringLength 60
#define cmToK   1.438769


#ifdef __cplusplus
extern "C"
#endif


void radtrans_(int *ok, int NumberMoleculeRadTrans, int querylength,                        // status variable, number of entries, length of query
               char queryString1[], char queryString2[],                                    // first and second part of query string
               int NumberOfMolecules,                                                       // number of molecules which should be considered
               char dbName[],                                                               // name of database
               char NameOfColumnName[],                                                     // name of column containing the name
               char MoleculeNames[NumberOfMolecules][nameLength],                           // name of molecules which should be considered
               char MolNameRadTrans[NumberMoleculeRadTrans][nameLength],                    // names of molecules
               double lFreq[NumberMoleculeRadTrans],                                        // frequency of the transition in MHz
               double lFreqErr[NumberMoleculeRadTrans],                                     // corresponding uncertainty in MHz
               double icatMol[NumberMoleculeRadTrans],                                      // base 10 logarithm of the integrated intensity
               double EinsteinA[NumberMoleculeRadTrans],                                    // Einstein A coefficient
               double lElt[NumberMoleculeRadTrans],                                         // energy of lower level
               int UpperStateDegeneracy[NumberMoleculeRadTrans],                            // upper state degeneracy
               char LowerQNString[NumberOfMolecules][QNStringLength],                       // quantum number string of lower state
               char UpperQNString[NumberOfMolecules][QNStringLength]                        // quantum number string of upper state
               )
{
    sqlite3 *db;                                                                            // sqlite3 db struct
    sqlite3_stmt *stmt;
    char *ErrorMsg = NULL;
    int rc;
    unsigned int col;                                                                       // loop variable

    char queryString[querylength];                                                          // final query string
    char querycopy[querylength];                                                            // copy of query string
    int i;                                                                                  // index variable
    int j;                                                                                  // loop variable
    int debug;                                                                              // used for debugging
    double pow();                                                                           // used for power function
    double val;                                                                             // working variable
    double valtemp;                                                                         // working variable

    debug = 0;                                                                              // set equal to 1 for debugging


    //----------------------------------------------------------------------------------------------------------------------------------------------------
    // construct final query string
    // SELECT * FROM ALMADB.RadTrans...

    // Debug:
    if (debug == 10){
        fprintf (stderr, "\n\n\nNumberMoleculeRadTrans = %i\n", NumberMoleculeRadTrans);
        fprintf (stderr, "querylength = %d\n", querylength);
        fprintf (stderr, "queryString1 = '%s'\n", queryString1);
        fprintf (stderr, "queryString2 = '%s'\n", queryString2);
        fprintf (stderr, "NumberOfMolecules = %i\n", NumberOfMolecules);
        fprintf (stderr, "dbName = '%s'\n", dbName);
        fprintf (stderr, "NameOfColumnName = '%s'\n", NameOfColumnName);
        fprintf (stderr, "\n\n");
        for (j = 0; j < NumberOfMolecules; j++)
        {
            fprintf (stderr, "i, MoleculeNames[j] = %i, '%s' \n", j, MoleculeNames[j]);
        }
        fprintf (stderr,"\n\n");
        for (j = 0; j < NumberMoleculeRadTrans; j++)
        {
            fprintf (stderr, "j = %i \n", j);
            fprintf (stderr, "MolNameRadTrans[j] = '%s' \n", MolNameRadTrans[j]);
            fprintf (stderr, "lFreq[j] = %f \n", lFreq[j]);
            fprintf (stderr, "lFreqErr[j] = %f \n", lFreqErr[j]);
            fprintf (stderr, "icatMol[j] = %f \n", icatMol[j]);
            fprintf (stderr, "EinsteinA[j] = %f \n", EinsteinA[j]);
            fprintf (stderr, "lElt[j] = %f \n", lElt[j]);
            fprintf (stderr, "UpperStateDegeneracy[j] = %i \n\n", UpperStateDegeneracy[j]);
            fprintf (stderr, "LowerQNString[j] = '%s' \n", LowerQNString[j]);
            fprintf (stderr, "UpperQNString[j] = '%s' \n", UpperQNString[j]);
        }
    }


    //----------------------------------------------------------------------------------------------------------------------------------------------------
    // connect to sqlite3 database
    rc = sqlite3_open(dbName, &db);
    if ( rc ){
        fprintf (stderr, "\n\n");
        fprintf (stderr, "subroutine numentr_sqlite3: Failed to connect to database: Error: %s\n", sqlite3_errmsg(db));
        fprintf (stderr, "\n\nqueryString = %s\n", queryString);
        *ok = 1;
        sqlite3_close(db);

        // Debug:
        if (debug == 1){fprintf (stderr,"rc = %i", rc);}


        /* check to see that the backend connection was successfully made */
    } else{

        // Debug:
        if (debug == 1){
            fprintf (stderr, "subroutine radtrans: Connection to database '%s' succeeded.\n", dbName);
            fprintf (stderr, "rc = %i\n", rc);
        }
        *ok = 0;
    }


    //----------------------------------------------------------------------------------------------------------------------------------------------------
    // collect number of entries for each molecule, even if a molecule occurs more than once
    i = 0;
    if (NumberOfMolecules > 0){
        if (debug == 1){fprintf (stderr, "\nNumberOfMolecules = %i\n", NumberOfMolecules);}


        //------------------------------------------------------------------------------------------------------------------------------------------------
        // loop over all molecules
        for (j = 0; j < NumberOfMolecules; j++){
            if (debug == 1){fprintf (stderr, "\n\nj = %i\n", j);}


            //--------------------------------------------------------------------------------------------------------------------------------------------
            // construct query string for each molecule
            sprintf(queryString, "%s WHERE (%s = %s", queryString1, NameOfColumnName, MoleculeNames[j]);
            sprintf(queryString, "%s) %s", queryString, queryString2);

            // Debug:
            if (debug == 1){fprintf (stderr, "\nfinal queryString = '%s'\n", queryString);}


            //--------------------------------------------------------------------------------------------------------------------------------------------
            // execute query
            rc = sqlite3_prepare_v2(db, queryString, -1, &stmt, NULL);
            if ( rc ){
                if (ErrorMsg != NULL){fprintf (stderr, "SQL error: %s", ErrorMsg);}
                sqlite3_free(ErrorMsg);
                ErrorMsg=NULL;
                *ok = 1;
                sqlite3_close(db);

                // Debug:
                if (debug == 1){fprintf (stderr, "rc = %i\n", rc);}
            }


            // get number of columns
            int cols = sqlite3_column_count(stmt);                                          // Read the number of rows fetched

            // Debug:
            if (debug == 1){fprintf (stderr, "\n\ncols = %i\n", cols);}


            //--------------------------------------------------------------------------------------------------------------------------------------------
            // get number of entries
            while(1)
            {
                // fetch a row’s status
                rc = sqlite3_step(stmt);
                if (rc == SQLITE_ROW)
                {
                    // sqlite3_column_text returns a const void* , typecast it to const char*
                    for (col = 0; col < cols; col++)
                    {
                        if (debug == 1){fprintf (stderr, "\ncol = %i\n", col);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get name of molecule
                        if (col == 0)
                        {
                            const char *text = (const char*)sqlite3_column_text(stmt, col);
                            sprintf(MolNameRadTrans[i], "%s", text);

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "name of molecule = '%s' \n", MolNameRadTrans[i]);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get (FREQ): Frequency of line in MHz
                        } else if (col == 1)
                        {
                            valtemp = sqlite3_column_double(stmt, col);
                            lFreq[i] = valtemp;

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "lFreq[i] = %f \n", lFreq[i]);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get (ERR): Estimated or experimental error of FREQ in MHz.
                        } else if (col == 2)
                        {
                            valtemp = sqlite3_column_double(stmt, col);
                            lFreqErr[i] = valtemp;

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "lFreqErr[i] = %f \n", lFreqErr[i]);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get (LGINT): Base 10 logarithm of the integrated intensity in units of nm^2 MHz at 300 K
                        }else if (col == 3)
                        {
                            valtemp = sqlite3_column_double(stmt, col);
                            val = pow(10., valtemp);
                            icatMol[i] = val;

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "icatMol[i] = %f \n", icatMol[i]);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get Einstein A coefficient
                        } else if (col == 4)
                        {
                            val = sqlite3_column_double(stmt, col);
                            EinsteinA[i] = val;

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "EinsteinA[i] = %f \n", EinsteinA[i]);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get (ELO): Lower state energy in cm^{-1} relative to the ground state
                        }else if (col == 5)
                        {
                            valtemp = sqlite3_column_double(stmt, col);
                            lElt[i] = valtemp * cmToK;

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "lElt[i] = %f \n", lElt[i]);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get (GUP): Upper state degeneracy
                        } else if (col == 6)
                        {
                            UpperStateDegeneracy[i] = sqlite3_column_int(stmt, col);

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "Upper state degeneracy = %i \n", UpperStateDegeneracy[i]);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get lower state quantum numbers
                        } else if (col == 7)
                        {
                            const char *text = (const char*)sqlite3_column_text(stmt, col);
                            strncpy(LowerQNString[i], text, QNStringLength - 1);
                            //LowerQNString[i][QNStringLength] = '\0';

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "lower state quantum numbers = '%s' \n", LowerQNString[i]);}


                        //--------------------------------------------------------------------------------------------------------------------------------
                        // get upper state quantum numbers
                        } else if (col == 8)
                        {
                            const char *text = (const char*)sqlite3_column_text(stmt, col);
                            strncpy(UpperQNString[i], text, QNStringLength - 1);
                            //UpperQNString[i][QNStringLength] = '\0';

                            // Debug:
                            if (debug >= 1){fprintf (stderr, "upper state quantum numbers = '%s' \n", UpperQNString[i]);}
                        }
                    }


                    // increase counter
                    i++;
                }
                else if(rc == SQLITE_DONE)
                {
                    break;
                }
            }


            //--------------------------------------------------------------------------------------------------------------------------------------------
            // finalize
            sqlite3_finalize(stmt);
        }
    }


    //----------------------------------------------------------------------------------------------------------------------------------------------------
    // close connection to database
    sqlite3_close(db);                                                                      // close the database connection

    // Debug:
    if (debug == 1){fprintf (stderr, "DONE \n");}
}

